searchNodes=[{"doc":"Computational flows with stages. Flow allows developers to express computations on collections, similar to the Enum and Stream modules, although computations will be executed in parallel using multiple GenStage s. Flow is designed to work with both bounded (finite) and unbounded (infinite) data. By default, Flow will work with batches of 500 items. This means Flow will only show improvements when working with larger collections. However, for certain cases, such as IO-bound flows, a smaller batch size can be configured through the :min_demand and :max_demand options supported by from_enumerable/2 , from_stages/2 , from_specs/2 , partition/2 , departition/5 , etc. Flow also provides the concepts of &quot;windows&quot; and &quot;triggers&quot;, which allow developers to split the data into arbitrary windows according to event time. Triggers allow computations to be materialized at different intervals, allowing developers to peek at results as they are computed. This module doc will cover the main constructs and concepts behind Flow, with examples. There is also a presentation about GenStage and Flow from José Valim at ElixirConf 2016, which covers data processing concepts for those unfamiliar with the domain: https://youtu.be/srtMWzyqdp8?t=244 Example As an example, let's implement the classic word counting algorithm using Flow. The word counting program will receive one file and count how many times each word appears in the document. Using the Enum module it could be implemented as follows: File . stream! ( &quot;path/to/some/file&quot; ) |&gt; Enum . flat_map ( &amp; String . split ( &amp;1 , &quot; &quot; ) ) |&gt; Enum . reduce ( %{ } , fn word , acc -&gt; Map . update ( acc , word , 1 , &amp; &amp;1 + 1 ) end ) |&gt; Enum . to_list ( ) Unfortunately, the implementation above is not very efficient, as Enum.flat_map/2 will build a list with all the words in the document before reducing it. If the document is, for example, 2GB, we will load 2GB of data into memory. We can improve the solution above by using the Stream module: File . stream! ( &quot;path/to/some/file&quot; ) |&gt; Stream . flat_map ( &amp; String . split ( &amp;1 , &quot; &quot; ) ) |&gt; Enum . reduce ( %{ } , fn word , acc -&gt; Map . update ( acc , word , 1 , &amp; &amp;1 + 1 ) end ) |&gt; Enum . to_list ( ) Now instead of loading the whole set into memory, we will only keep the current line in memory while we process it. While this allows us to process the whole data set efficiently, it does not leverage concurrency. Flow solves that: File . stream! ( &quot;path/to/some/file&quot; ) |&gt; Flow . from_enumerable ( ) |&gt; Flow . flat_map ( &amp; String . split ( &amp;1 , &quot; &quot; ) ) |&gt; Flow . partition ( ) |&gt; Flow . reduce ( fn -&gt; %{ } end , fn word , acc -&gt; Map . update ( acc , word , 1 , &amp; &amp;1 + 1 ) end ) |&gt; Enum . to_list ( ) To convert from Stream to Flow, we have made two changes: We have replaced the calls to Stream with Flow We call partition/2 so words are properly partitioned between stages The example above will use all available cores and will keep an ongoing flow of data instead of traversing them line by line. Once all data is computed, it is sent to the process which invoked Enum.to_list/1 . While we gain concurrency by using Flow, many of the benefits of Flow are in partitioning the data. We will discuss the need for data partitioning next. Partitioning To understand the need to partition the data, let's change the example above and remove the partition call: File . stream! ( &quot;path/to/some/file&quot; ) |&gt; Flow . from_enumerable ( ) |&gt; Flow . flat_map ( &amp; String . split ( &amp;1 , &quot; &quot; ) ) |&gt; Flow . reduce ( fn -&gt; %{ } end , fn word , acc -&gt; Map . update ( acc , word , 1 , &amp; &amp;1 + 1 ) end ) |&gt; Enum . to_list ( ) This will execute the flat_map and reduce operations in parallel inside multiple stages. When running on a machine with two cores: [ file stream ] # Flow.from_enumerable/1 (producer) | | [ M1 ] [ M2 ] # Flow.flat_map/2 + Flow.reduce/3 (consumer) Now imagine that the M1 and M2 stages above receive the following lines: M1 - &quot;roses are red&quot; M2 - &quot;violets are blue&quot; flat_map/2 will break them into: M1 - [ &quot;roses&quot; , &quot;are&quot; , &quot;red&quot; ] M2 - [ &quot;violets&quot; , &quot;are&quot; , &quot;blue&quot; ] Then reduce/3 will result in each stage having the following state: M1 - %{ &quot;roses&quot; =&gt; 1 , &quot;are&quot; =&gt; 1 , &quot;red&quot; =&gt; 1 } M2 - %{ &quot;violets&quot; =&gt; 1 , &quot;are&quot; =&gt; 1 , &quot;blue&quot; =&gt; 1 } Which is converted to the list (in no particular order): [ { &quot;roses&quot; , 1 } , { &quot;are&quot; , 1 } , { &quot;red&quot; , 1 } , { &quot;violets&quot; , 1 } , { &quot;are&quot; , 1 } , { &quot;blue&quot; , 1 } ] Although both stages have performed word counting, we have words like &quot;are&quot; that appear on both stages. This means we would need to perform yet another pass on the data merging the duplicated words across stages. This step would have to run on a single process, which would limit our ability to run concurrently. Remember that events are batched, so for small files, there is a chance all lines will be set to the same stage (M1 or M2) and you won't be able to replicate the issue. If you want to emulate this, either to follow along or in your test suites, you may set :max_demand to 1 when reading from the stream, so that the code looks like this: File . stream! ( &quot;path/to/some/file&quot; ) |&gt; Flow . from_enumerable ( max_demand : 1 ) |&gt; Flow . flat_map ( &amp; String . split ( &amp;1 , &quot; &quot; ) ) |&gt; Flow . reduce ( fn -&gt; %{ } end , fn word , acc -&gt; Map . update ( acc , word , 1 , &amp; &amp;1 + 1 ) end ) |&gt; Enum . to_list ( ) Partitioning solves this by introducing a new set of stages and making sure the same word is always mapped to the same stage with the help of a hash function. Let's introduce the call to partition/2 back: File . stream! ( &quot;path/to/some/file&quot; ) |&gt; Flow . from_enumerable ( ) |&gt; Flow . flat_map ( &amp; String . split ( &amp;1 , &quot; &quot; ) ) |&gt; Flow . partition ( ) |&gt; Flow . reduce ( fn -&gt; %{ } end , fn word , acc -&gt; Map . update ( acc , word , 1 , &amp; &amp;1 + 1 ) end ) |&gt; Enum . to_list ( ) Now we will have the following topology: [ file stream ] # Flow.from_enumerable/1 (producer) | | [ M1 ] [ M2 ] # Flow.flat_map/2 (producer-consumer) | \\ / | | \\ / | | / \\ | [ R1 ] [ R2 ] # Flow.reduce/3 (consumer) If the M1 and M2 stages receive the same lines and break them into words as before: M1 - [ &quot;roses&quot; , &quot;are&quot; , &quot;red&quot; ] M2 - [ &quot;violets&quot; , &quot;are&quot; , &quot;blue&quot; ] Now, any given word will be consistently routed to R1 or R2 regardless of its origin. The default hashing function will route them like this: R1 - [ &quot;roses&quot; , &quot;are&quot; , &quot;red&quot; , &quot;are&quot; ] R2 - [ &quot;violets&quot; , &quot;blue&quot; ] Resulting in the reduced state of: R1 - %{ &quot;roses&quot; =&gt; 1 , &quot;are&quot; =&gt; 2 , &quot;red&quot; =&gt; 1 } R2 - %{ &quot;violets&quot; =&gt; 1 , &quot;blue&quot; =&gt; 1 } Which is converted to the list (in no particular order): [ { &quot;roses&quot; , 1 } , { &quot;are&quot; , 2 } , { &quot;red&quot; , 1 } , { &quot;violets&quot; , 1 } , { &quot;blue&quot; , 1 } ] Each stage has a distinct subset of the data so we know that we don't need to merge the data later on, because a given word is guaranteed to have only been routed to one stage. Partitioning the data is a very useful technique. For example, if we wanted to count the number of unique elements in a dataset, we could perform such a count in each partition and then sum their results, as the partitioning guarantees the data in each partition won't overlap. A unique element would never be counted twice. The topology above alongside partitioning is very common in the MapReduce programming model which we will briefly discuss next. MapReduce The MapReduce programming model forces us to break our computations in two stages: map and reduce. The map stage is often quite easy to parallelize because events are processed individually and in isolation. The reduce stages need to group the data either partially or completely. In the example above, the stages executing flat_map/2 are the mapper stages. Because the flat_map/2 function works line by line, we can have two, four, eight or more mapper processes that will break line by line into words without any need for coordination. However, the reducing stage is a bit more complicated. Reducer stages typically aggregate some result based on their inputs, such as how many times a word has appeared. This implies reducer computations need to traverse the whole data set and, in order to do so in parallel, we partition the data into distinct datasets. The goal of the reduce/3 operation is to accumulate a value which then becomes the partition state. Any operation that happens after reduce/3 works on the whole state and is only executed after all the data for a partition is collected. While this approach works well for bounded (finite) data, it is quite limited for unbounded (infinite) data. After all, if the reduce operation needs to traverse the whole partition to complete, how can we do so if the data never finishes? The answer here lies in triggers. Every partition may have a on_trigger/2 callback which receives the partition accumulator and returns the events to be emitted and the accumulator to be used after the trigger. All flows have at least one trigger: the :done trigger which is executed when all the data has been processed. In this case, the accumulator returned by on_trigger/2 won't be used, only the events it emits. However, Flow provides many conveniences for working with unbound data, allowing us to set windows, time-based triggers, element counters and more. Data completion, windows and triggers By default, Flow shuts down its processes when all data has been processed. However, when working with an unbounded stream of data, there is no such thing as data completion. So when can we consider a reduce function to be &quot;completed&quot;? To handle such cases, Flow provides windows and triggers. Windows allow us to split the data based on the event time while triggers tells us when to write the results we have computed so far. By introducing windows, we no longer think about events being partitioned across stages. Instead each event belongs to a window and the window is partitioned across the stages. By default, all events belong to the same window (called the global window), which is partitioned across stages. However, different windowing strategies can be used by building a Flow.Window and passing it to the Flow.partition/2 function. Once a window is specified, we can create triggers that tell us when to checkpoint the data, allowing us to report our progress while the data streams through the system, regardless of whether the data is bounded or unbounded. Every time a trigger is invoked, the on_trigger/2 callback of that partition is invoked, allowing us to control which events to emit and what accumulator to use for the next time the partition starts reducing data. Windows and triggers effectively control how the reduce/3 function works. While windows and triggers allow us to control when data is emitted, note that data can be emitted at any time during the reducing step by using emit_and_reduce/3 . In truth, all window and trigger functionality provided by Flow can also be built by hand using the emit_and_reduce/3 and on_trigger/2 functions. See Flow.Window for a complete introduction to windows and triggers. Supervisable flows In the examples so far we have started a flow dynamically and consumed it using Enum.to_list/1 . Unfortunately calling a function from Enum will cause the whole computed dataset to be sent to a single process. In many situations, this is either too expensive or completely undesirable. For example, in data-processing pipelines, it is common to receive data continuously from external sources. At the end, this data is written to disk or another storage mechanism after being processed, rather than being sent to a single process. Flow allows computations to be started as a group of processes which may run indefinitely. This can be done by starting the flow as part of a supervision tree using {Flow, your_flow} as your child specification: children = [ { Flow , Flow . from_stages ( ... ) |&gt; Flow . flat_map ( &amp; String . split ( &amp;1 , &quot; &quot; ) ) |&gt; Flow . reduce ( fn -&gt; %{ } end , fn word , acc -&gt; Map . update ( acc , word , 1 , &amp; &amp;1 + 1 ) end ) } ] It is also possible to move a Flow to its own module. This is done by calling use Flow and then defining a start_link/1 function that calls Flow.start_link/1 at the end: defmodule MyFlow do use Flow def start_link ( _ ) do Flow . from_stages ( ... ) |&gt; Flow . flat_map ( &amp; String . split ( &amp;1 , &quot; &quot; ) ) |&gt; Flow . reduce ( fn -&gt; %{ } end , fn word , acc -&gt; Map . update ( acc , word , 1 , &amp; &amp;1 + 1 ) end ) |&gt; Flow . start_link ( ) end end By the default the Flow is permanent, which means it is always restarted. The :shutdown and :restart child spec configurations can be given to use Flow . Flow also provides integration with GenStage , allowing you to specify child specifications of producers, producer consumers, and consumers that are started alongside the flow and under the same supervision tree. This is achieved with the from_specs/2 (producers), through_specs/2 (producer consumers) and into_specs/2 (consumers) functions. It is also possible to connect a flow to already running stages, via the from_stages/2 (producers), through_stages/2 (producer consumers) and into_stages/2 (consumers) functions. into_stages/3 and into_specs/3 are alternatives to start_link/1 that start the flow with the given consumers stages or the given consumers child specification. Similar to start_link/1 , they return either {:ok, pid} or {:error, reason} . Performance discussions In this section we will discuss points related to performance with flows. Know your code There are many optimizations we could perform in the flow above that are not necessarily related to flows themselves. Let's rewrite the flow using some of them: # The parent process which will own the table parent = self ( ) # Let&#39;s compile common patterns for performance empty_space = :binary . compile_pattern ( &quot; &quot; ) # BINARY File . stream! ( &quot;path/to/some/file&quot; , read_ahead : 100_000 ) # READ_AHEAD |&gt; Flow . from_enumerable ( ) |&gt; Flow . flat_map ( &amp; String . split ( &amp;1 , empty_space ) ) # BINARY |&gt; Flow . partition ( ) |&gt; Flow . reduce ( fn -&gt; :ets . new ( :words , [ ] ) end , fn word , ets -&gt; # ETS :ets . update_counter ( ets , word , { 2 , 1 } , { word , 0 } ) ets end ) |&gt; Flow . on_trigger ( fn ets -&gt; :ets . give_away ( ets , parent , [ ] ) { [ ets ] , :new_reduce_state_which_wont_be_used } # Emit the ETS end ) |&gt; Enum . to_list ( ) We have performed three optimizations: BINARY - the first optimization is to compile the pattern we use to split the string on READ_AHEAD - the second optimization is to use the :read_ahead option for file streams allowing us to do fewer IO operations by reading large chunks of data at once ETS - the third stores the data in a ETS table and uses its counter operations. For counters and a large dataset this provides a great performance benefit as it generates less garbage. At the end, we call on_trigger/2 to transfer the ETS table to the parent process and wrap the table in a list so we can access it on Enum.to_list/1 . This step is not strictly required. For example, one could write the table to disk with :ets.tab2file/2 at the end of the computation Configuration (demand and the number of stages) from_enumerable/2 , from_stages/2 and partition/3 allow a set of options to configure how flows work. In particular, we recommend that developers play with the :min_demand and :max_demand options, which control the amount of data sent between stages. The difference between max_demand and min_demand works as the batch size when the producer is full. If the producer has fewer events than requested by consumers, it usually sends the remaining events available. If stages perform IO, it may also be worth increasing the number of stages. The default value is System.schedulers_online/0 , which is a good default if the stages are CPU bound, but if stages are waiting on external resources or other processes, increasing the number of stages may be helpful. Avoid single sources In the examples so far we have used a single file as our data source. In practice such single sources should be avoided as they could end up being the bottleneck of our whole computation. In the file stream case above, instead of having one single large file, it is preferable to break the file into smaller ones: streams = for file &lt;- File . ls! ( &quot;dir/with/files&quot; ) do File . stream! ( &quot;dir/with/files/ \#{ file } &quot; , read_ahead : 100_000 ) end streams |&gt; Flow . from_enumerables ( ) |&gt; Flow . flat_map ( &amp; String . split ( &amp;1 , &quot; &quot; ) ) |&gt; Flow . partition ( ) |&gt; Flow . reduce ( fn -&gt; %{ } end , fn word , acc -&gt; Map . update ( acc , word , 1 , &amp; &amp;1 + 1 ) end ) |&gt; Enum . to_list ( ) Instead of calling from_enumerable/1 , we now called from_enumerables/1 which expects a list of enumerables to be used as source. Notice every stream also uses the :read_ahead option which tells Elixir to buffer file data in memory to avoid multiple IO lookups. If the number of enumerables is equal to or greater than the number of cores, Flow will automatically fuse the enumerables with the mapper logic. For example, if three file streams are given as enumerables to a machine with two cores, we will have the following topology: [ F1 ] [ F2 ] [ F3 ] # file stream [ M1 ] [ M2 ] [ M3 ] # Flow.flat_map/2 (producer) | \\ / \\ / | | / \\ / \\ | | // \\\\ | [ R1 ] [ R2 ] # Flow.reduce/3 (consumer)","ref":"Flow.html","title":"Flow","type":"module"},{"doc":"Joins two bounded (finite) flows. It expects the left and right flow, the left_key and right_key to calculate the key for both flows and the join function which is invoked whenever there is a match. A join creates a new partitioned flow that subscribes to the two flows given as arguments. The newly created partitions will accumulate the data received from both flows until there is no more data. Therefore, this function is useful for merging finite flows. If used for merging infinite flows, you will eventually run out of memory due to the accumulated data. See window_join/8 for applying a window to a join, allowing the join data to be reset per window. The join has 4 modes: :inner - data will only be emitted when there is a match between the keys in left and right side :left_outer - similar to :inner plus all items given in the left that did not have a match will be emitted at the end with nil for the right value :right_outer - similar to :inner plus all items given in the right that did not have a match will be emitted at the end with nil for the left value :full_outer - similar to :inner plus all items given in the left and right that did not have a match will be emitted at the end with nil for the right and left value respectively The joined partitions can be configured via options with the same values as shown on from_enumerable/2 or from_stages/2 . Examples iex&gt; posts = [ %{ id : 1 , title : &quot;hello&quot; } , %{ id : 2 , title : &quot;world&quot; } ] iex&gt; comments = [ { 1 , &quot;excellent&quot; } , { 1 , &quot;outstanding&quot; } , ...&gt; { 2 , &quot;great follow up&quot; } , { 3 , &quot;unknown&quot; } ] iex&gt; flow = Flow . bounded_join ( :inner , ...&gt; Flow . from_enumerable ( posts ) , ...&gt; Flow . from_enumerable ( comments ) , ...&gt; &amp; &amp;1 . id , # left key ...&gt; &amp; elem ( &amp;1 , 0 ) , # right key ...&gt; fn post , { _post_id , comment } -&gt; Map . put ( post , :comment , comment ) end ) iex&gt; Enum . sort ( flow ) [ %{ id : 1 , title : &quot;hello&quot; , comment : &quot;excellent&quot; } , %{ id : 2 , title : &quot;world&quot; , comment : &quot;great follow up&quot; } , %{ id : 1 , title : &quot;hello&quot; , comment : &quot;outstanding&quot; } ]","ref":"Flow.html#bounded_join/7","title":"Flow.bounded_join/7","type":"function"},{"doc":"Reduces windows over multiple partitions into a single stage. Once departition/5 is called, computations no longer happen concurrently until the data is once again partitioned. departition/5 is typically invoked as the last step in a flow to merge the state from all previous partitions per window. It requires a flow and three functions as arguments as described: the accumulator function - a zero-arity function that returns the initial accumulator. This function is invoked per window. the merger function - a function that receives the state of a given partition and the accumulator and merges them together. the done function - a function that receives the final accumulator. A set of options may also be given to customize the :window , :min_demand and :max_demand . Examples For example, imagine we are counting words in a document. Each partition ends up with a map of words as keys and count as values. In the examples in the module documentation, we streamed those results to a single client using Enum.to_list/1 . However, we could use departition/5 to reduce the data over multiple stages returning one single map with all results: File . stream! ( &quot;path/to/some/file&quot; ) |&gt; Flow . from_enumerable ( ) |&gt; Flow . map ( &amp; String . split / 1 ) |&gt; Flow . partition ( ) |&gt; Flow . reduce ( fn -&gt; %{ } end , fn event , acc -&gt; Map . update ( acc , event , 1 , &amp; &amp;1 + 1 ) end ) |&gt; Flow . departition ( &amp; Map . new / 0 , &amp; Map . merge / 2 , &amp; ( &amp;1 ) ) |&gt; Enum . to_list The departition function expects the initial accumulator, a function that merges the data, and a final function invoked when the computation is done. Departition also works with windows and triggers. A new accumulator is created per window and the merge function is invoked with the state every time a trigger is emitted in any of the partitions. This can be useful to compute the final state as computations happen instead of one time at the end. For example, we could change the flow above so each partition emits their whole intermediary state every 1000 items, merging it into the departition more frequently: File . stream! ( &quot;path/to/some/file&quot; ) |&gt; Flow . from_enumerable ( ) |&gt; Flow . map ( &amp; String . split / 1 ) |&gt; Flow . partition ( window : Flow.Window . global |&gt; Flow.Window . trigger_every ( 1000 ) ) |&gt; Flow . reduce ( fn -&gt; %{ } end , fn event , acc -&gt; Map . update ( acc , event , 1 , &amp; &amp;1 + 1 ) end ) |&gt; Flow . on_trigger ( fn acc -&gt; { [ acc ] , %{ } } end ) |&gt; Flow . departition ( &amp; Map . new / 0 , &amp; Map . merge ( &amp;1 , &amp;2 , fn _ , v1 , v2 -&gt; v1 + v2 end ) , &amp; ( &amp;1 ) ) |&gt; Enum . to_list Each approach is going to have different performance characteristics and it is important to measure to verify which one will be more efficient to the problem at hand.","ref":"Flow.html#departition/5","title":"Flow.departition/5","type":"function"},{"doc":"","ref":"Flow.html#each/2","title":"Flow.each/2","type":"function"},{"doc":"Controls which values should be emitted. The argument can be either :events , :state or :nothing . This step must be called after the reduce operation and it will guarantee the state is a list that can be sent downstream. Most commonly :events is used and each partition will emit the events it has processed to the next stages. However, sometimes we want to emit counters or other data structures as a result of our computations. In such cases, the emit argument can be set to :state , to return the :state from reduce/3 or even the processed collection as a whole.","ref":"Flow.html#emit/2","title":"Flow.emit/2","type":"function"},{"doc":"Reduces values with the given accumulator and controls which values should be emitted. acc_fun is a function that receives no arguments and returns the actual accumulator. The acc_fun function is invoked per window whenever a new window starts. If a trigger is emitted and it is configured to reset the accumulator, the acc_fun function will be invoked once again. This function behaves similarly to reduce/3 , but in addition to accumulating data, it also gives full control over what will be emitted. reducer_fun must return a tuple where the first element is the list of events to be emitted and the second is the new state of the accumulator. Examples As an example this is a simple implementation of a sliding window of 3 events. The reducer function always emits a list of the most recent (at most) 3 events. Note that at the end of the input the current state of the accumulator will be emitted which we filter in this example at the last step. iex&gt; flow = Flow . from_enumerable ( 1 .. 5 , stages : 1 ) iex&gt; flow = flow |&gt; Flow . emit_and_reduce ( fn -&gt; [ ] end , fn event , acc -&gt; ...&gt; acc = [ event | acc ] |&gt; Enum . take ( 3 ) ...&gt; { [ Enum . reverse ( acc ) ] , acc } ...&gt; end ) iex&gt; flow |&gt; Enum . filter ( &amp; is_list / 1 ) [ [ 1 ] , [ 1 , 2 ] , [ 1 , 2 , 3 ] , [ 2 , 3 , 4 ] , [ 3 , 4 , 5 ] ]","ref":"Flow.html#emit_and_reduce/3","title":"Flow.emit_and_reduce/3","type":"function"},{"doc":"Applies the given function filtering each input in parallel. Examples iex&gt; flow = [ 1 , 2 , 3 ] |&gt; Flow . from_enumerable ( ) |&gt; Flow . filter ( &amp; ( rem ( &amp;1 , 2 ) == 0 ) ) iex&gt; Enum . sort ( flow ) # Call sort as we have no order guarantee [ 2 ]","ref":"Flow.html#filter/2","title":"Flow.filter/2","type":"function"},{"doc":"Applies the given function mapping each input in parallel and flattening the result, but only one level deep. Examples iex&gt; flow = [ 1 , 2 , 3 ] |&gt; Flow . from_enumerable ( ) |&gt; Flow . flat_map ( fn x -&gt; [ x , x * 2 ] end ) iex&gt; Enum . sort ( flow ) # Call sort as we have no order guarantee [ 1 , 2 , 2 , 3 , 4 , 6 ]","ref":"Flow.html#flat_map/2","title":"Flow.flat_map/2","type":"function"},{"doc":"Creates a flow with the given enumerable as the producer. Calling this function is equivalent to: Flow . from_enumerables ( [ enumerable ] , options ) The enumerable is consumed in batches, retrieving max_demand items the first time and then max_demand - min_demand the next times. Therefore, for streams that cannot produce items that fast, it is recommended to pass a lower :max_demand value as an option. It is also expected the enumerable is able to produce the whole batch on demand or terminate. If the enumerable is a blocking one, for example, because it needs to wait for data from another source, it will block until the current batch is fully filled. GenStage and Flow were created exactly to address such issue. So if you have a blocking enumerable that you want to use in your Flow, then it must be implemented with GenStage and integrated with from_stages/2 . Examples &quot;some/file&quot; |&gt; File . stream! ( read_ahead : 100_000 ) |&gt; Flow . from_enumerable ( ) some_network_based_stream ( ) |&gt; Flow . from_enumerable ( max_demand : 20 )","ref":"Flow.html#from_enumerable/2","title":"Flow.from_enumerable/2","type":"function"},{"doc":"Creates a flow with the given enumerable as producer. The enumerable is consumed in batches, retrieving max_demand items the first time and then max_demand - min_demand the next times. Therefore, for streams that cannot produce items that fast, it is recommended to pass a lower :max_demand value as an option. See GenStage.from_enumerable/2 for information and limitations on enumerable-based stages. Options These options configure the stages connected to producers before partitioning. :window - a window to run the next stages in, see Flow.Window :stages - the number of stages :buffer_keep - how the buffer should behave, see GenStage.init/1 :buffer_size - how many events to buffer, see GenStage.init/1 :shutdown - the shutdown time for this stage when the flow is shut down. The same as the :shutdown value in a Supervisor, defaults to 5000 milliseconds. :on_init - a function invoked during the initialization of each stage. The function receives a single argument in the form of {i, total} where: i is the stage index total is the total number of stages All remaining options are sent during subscription, allowing developers to customize :min_demand , :max_demand and others. Examples files = [ File . stream! ( &quot;some/file1&quot; , read_ahead : 100_000 ) , File . stream! ( &quot;some/file2&quot; , read_ahead : 100_000 ) , File . stream! ( &quot;some/file3&quot; , read_ahead : 100_000 ) ] Flow . from_enumerables ( files )","ref":"Flow.html#from_enumerables/2","title":"Flow.from_enumerables/2","type":"function"},{"doc":"Creates a flow with a list of producers child specifications. The child specification is the one defined in the Supervisor module. The producers will only be started when the flow starts. If the flow terminates, the producers will also be terminated. The :id field of the child specification will be randomized. The :restart option is set to :temporary but it behaves as :transient . If a producer terminates, its exit reason will propagate through the flow. The exit is considered abnormal unless the reason is :normal , :shutdown or {:shutdown, _} . All other child specification fields are kept unchanged. For options and termination behaviour, see from_stages/2 . Examples specs = [ { MyProducer , arg1 } , { MyProducer , arg2 } ] Flow . from_specs ( specs )","ref":"Flow.html#from_specs/2","title":"Flow.from_specs/2","type":"function"},{"doc":"Creates a flow with a list of already running stages as producers . producers are already running stages that have type :producer If instead you want the producers to be started alongside the flow, see from_specs/2 instead. Options These options configure the stages connected to producers before partitioning. :window - a window to run the next stages in, see Flow.Window :stages - the number of stages :buffer_keep - how the buffer should behave, see GenStage.init/1 :buffer_size - how many events to buffer, see GenStage.init/1 :shutdown - the shutdown time for this stage when the flow is shut down. The same as the :shutdown value in a Supervisor, defaults to 5000 milliseconds. All remaining options are sent during subscription, allowing developers to customize :min_demand , :max_demand and others. Examples stages = [ pid1 , pid2 , pid3 ] Flow . from_stages ( stages ) Termination Flow subscribes to producer stages using cancel: :transient . This means producer stages can signal the flow that it has emitted all events by terminating with reason :normal , :shutdown or {:shutdown, _} . Therefore, if you are implementing a producer that may eventually terminate, then the producer must exit with reason :normal , :shutdown or {:shutdown, _} after emitting all events. This is often done in the producer by using GenStage.async_info(self(), :terminate) to send a message to itself once all events have been dispatched: def handle_info ( :terminate , state ) do { :stop , :shutdown , state } end Once all producers have finished, the stages subscribed to the producer will terminate, causing the next layer of stages in the flow to terminate and so forth, until the whole flow shuts down. If the exit reason is none of the above, it will cause the next stages to terminate immediately, eventually causing the whole flow to terminate.","ref":"Flow.html#from_stages/2","title":"Flow.from_stages/2","type":"function"},{"doc":"Groups events with the given key_fun . This is a reduce operation that groups events into maps where the key is the key returned by key_fun and the value is a list of values in reverse order as returned by value_fun . The resulting map becomes the stage state. Examples iex&gt; flow = Flow . from_enumerable ( ~w[the quick brown fox] , stages : 1 ) iex&gt; flow |&gt; Flow . group_by ( &amp; String . length / 1 ) |&gt; Enum . sort ( ) [ { 3 , [ &quot;fox&quot; , &quot;the&quot; ] } , { 5 , [ &quot;brown&quot; , &quot;quick&quot; ] } ]","ref":"Flow.html#group_by/3","title":"Flow.group_by/3","type":"function"},{"doc":"Groups a series of {key, value} tuples by keys. This is a reduce operation that groups events into maps with the given key and a list of values with the given keys in reverse order. The resulting map becomes the stage state. Examples iex&gt; flow = Flow . from_enumerable ( [ foo : 1 , foo : 2 , bar : 3 , foo : 4 , bar : 5 ] , stages : 1 ) iex&gt; flow |&gt; Flow . group_by_key ( ) |&gt; Flow . emit ( :state ) |&gt; Enum . to_list ( ) [ %{ foo : [ 4 , 2 , 1 ] , bar : [ 5 , 3 ] } ]","ref":"Flow.html#group_by_key/1","title":"Flow.group_by_key/1","type":"function"},{"doc":"Starts a flow and the consumers child specifications. consumers is a list of tuples where the first element is the child specification and the second is a list of subscription options. The child specification is the one defined in the Supervisor module. The consumers will only be started when the flow starts. If the flow terminates, the consumers will also be terminated. The :id field of the child specification will be randomized. All other fields are kept as in. If the consumer terminates, it will behave according to its restart strategy. Once a consumer terminates, the whole flow is terminated. For options and termination behaviour, see into_stages/3 . Examples spec = { MyConsumer , arg } subscription_opts = [ ] specs = [ { spec , subscription_opts } ] Flow . into_specs ( some_flow , specs )","ref":"Flow.html#into_specs/3","title":"Flow.into_specs/3","type":"function"},{"doc":"Starts a flow with a list of already running stages as consumers . consumers is a list of already running stages that have type :consumer or :producer_consumer . Each element represents the consumer or a tuple with the consumer and the subscription options as defined in GenStage.sync_subscribe/2 . If instead you want the consumers to be started alongside the flow, see into_specs/3 instead. The pid returned by this function identifies a coordinator process. While it is possible to send subscribe requests to the coordinator process, the coordinator process will simply redirect the subscription to the proper flow processes and cancel the initial subscription. This means late subscriptions should use at cancel: :transient (which is the default for stage subscriptions). Keep in mind this implies consumers will continue running when the producers exits with :normal or :shutdown reason. The coordinator exits with reason :normal only if all consumers exit with reason :normal . Otherwise exits with reason :shutdown . Options This function receives the same options as start_link/2 with the addition of a :dispatcher option that configures how the consumers get data from the flow and defaults to GenStage.DemandDispatch . It may be either an atom or a tuple with the dispatcher and the dispatcher options. Termination Flow subscribes to stages using cancel: :transient . This means stages can signal the flow that it has emitted all events by terminating with reason :normal , :shutdown or {:shutdown, _} . If you are implementing your own consumer and you are subscribing to a flow that is finite, you need to take this into account in your consumer implementation if you want proper consumer termination: You need implement GenStage.handle_subscribe/4 and store whenever the stage gets a new producer You need implement GenStage.handle_cancel/3 and decrease whenever the stage loses a producer Once all producers are cancelled, you can terminate: def handle_info ( :terminate , state ) do { :stop , :shutdown , state } end Given the complexity in guaranteeing termination, we recommend developers to use into_stages/3 and into_specs/3 only when subscribing to unbounded (infinite) flows.","ref":"Flow.html#into_stages/3","title":"Flow.into_stages/3","type":"function"},{"doc":"Applies the given function mapping each input in parallel. Examples iex&gt; flow = [ 1 , 2 , 3 ] |&gt; Flow . from_enumerable ( ) |&gt; Flow . map ( &amp; ( &amp;1 * 2 ) ) iex&gt; Enum . sort ( flow ) # Call sort as we have no order guarantee [ 2 , 4 , 6 ] iex&gt; flow = Flow . from_enumerables ( [ [ 1 , 2 , 3 ] , 1 .. 3 ] ) |&gt; Flow . map ( &amp; ( &amp;1 * 2 ) ) iex&gt; Enum . sort ( flow ) [ 2 , 2 , 4 , 4 , 6 , 6 ]","ref":"Flow.html#map/2","title":"Flow.map/2","type":"function"},{"doc":"Applies the given function to each &quot;batch&quot; of GenStage events. Flow uses GenStage which sends events in batches, controlled by min_demand and max_demand . This callback allows you to hook into this batch, before any map or reduce operation is invoked. This often useful to preload data that is used in later stages.","ref":"Flow.html#map_batch/2","title":"Flow.map_batch/2","type":"function"},{"doc":"Maps over the given values in the stage state. It is expected the state to emit two-elements tuples, such as list, maps, etc. Examples iex&gt; flow = Flow . from_enumerable ( [ a : 1 , b : 2 , c : 3 , d : 4 , e : 5 ] , stages : 1 ) iex&gt; flow |&gt; Flow . map_values ( &amp; &amp;1 * 2 ) |&gt; Enum . sort ( ) [ a : 2 , b : 4 , c : 6 , d : 8 , e : 10 ]","ref":"Flow.html#map_values/2","title":"Flow.map_values/2","type":"function"},{"doc":"Merges the given flow or flows into a series of new stages with the given dispatcher and options. This is the function used as building block by partition/2 and shuffle/2 . Options :window - a Flow.Window struct which controls how the reducing function behaves, see Flow.Window for more information. :stages - the number of partitions (reducer stages) :shutdown - the shutdown time for this stage when the flow is shut down. The same as the :shutdown value in a Supervisor, defaults to 5000 milliseconds.","ref":"Flow.html#merge/3","title":"Flow.merge/3","type":"function"},{"doc":"Applies the given function over the window state. This function must be called after group_by/3 , reduce/3 or emit_and_reduce/3 as it works on the accumulated state. on_trigger/2 is invoked per window on every stage whenever there is a trigger: this gives us an understanding of the window data while leveraging the parallelism between stages. The given callback must return a tuple with elements to emit and the new accumulator. The callback arguments The callback function may have arity 1, 2 or 3. The first argument is the state. The second argument is optional and contains the partition index. The partition index is a two-element tuple identifying the current partition and the total number of partitions as the second element. For example, for a partition with 4 stages, the partition index will be the values {0, 4} , {1, 4} , {2, 4} and {3, 4} . The third argument is optional and contains the window-trigger information. This information is a three-element tuple containing the window name, the window identifier, and the trigger name. For example, a global window created with Flow.Window.global/0 will emit on termination: { :global , :global , :done } A Flow.Window.global/0 window with a count trigger created with Flow.Window.trigger_every/2 will also emit: { :global , :global , { :every , 20 } } A Flow.Window.fixed/3 window will emit on done: { :fixed , window , :done } Where window is an integer identifying the timestamp for the window being triggered. Examples We can use on_trigger/2 to transform the collection after processing. For example, if we want to count the amount of unique letters in a sentence, we can partition the data, then reduce over the unique entries and finally return the size of each stage, summing it all: iex&gt; flow = Flow . from_enumerable ( [ &quot;the quick brown fox&quot; ] ) |&gt; Flow . flat_map ( fn word -&gt; ...&gt; String . graphemes ( word ) ...&gt; end ) iex&gt; flow = Flow . partition ( flow ) iex&gt; flow = Flow . reduce ( flow , fn -&gt; %{ } end , &amp; Map . put ( &amp;2 , &amp;1 , true ) ) iex&gt; flow |&gt; Flow . on_trigger ( fn map -&gt; { [ map_size ( map ) ] , map } end ) |&gt; Enum . sum ( ) 16","ref":"Flow.html#on_trigger/2","title":"Flow.on_trigger/2","type":"function"},{"doc":"Creates a new partition for the given flow (or flows) with the given options. Every time this function is called, a new partition is created. It is typically recommended to invoke it before a reducing function, such as reduce/3 , so data belonging to the same partition can be kept together. However, notice that unnecessary partitioning will increase memory usage and reduce throughput with no benefit whatsoever. Flow takes care of using all cores regardless of the number of times you call partition. You should only partition when the problem you are trying to solve requires you to route the data around. Such as the problem presented in Flow 's module documentation. If you can solve a problem without using partition at all, that is typically preferred. Those are typically called &quot;embarrassingly parallel&quot; problems. Examples flow |&gt; Flow . partition ( window : Flow.Window . global ) flow |&gt; Flow . partition ( stages : 4 ) Options :window - a Flow.Window struct which controls how the reducing function behaves, see Flow.Window for more information. :stages - the number of partitions (reducer stages) :shutdown - the shutdown time for this stage when the flow is shut down. The same as the :shutdown value in a Supervisor, defaults to 5000 milliseconds. :key - the key to use when partitioning. It is a function that receives a single argument (the event) and must return its key. The key will then be hashed by Flow. To facilitate customization, :key also allows common values, such as {:elem, integer} and {:key, atom} , to calculate the hash based on a tuple or a map field. See the &quot;Key shortcuts&quot; section below :hash - the hashing function. By default a hashing function is built on the key but a custom one may be specified as described in GenStage.PartitionDispatcher :min_demand - the minimum demand for this subscription :max_demand - the maximum demand for this subscription Key shortcuts The following shortcuts can be given to the :key option: {:elem, index} - apply the hash function to the element at index (zero-based) in the given tuple {:key, key} - apply the hash function to the key of a given map","ref":"Flow.html#partition/2","title":"Flow.partition/2","type":"function"},{"doc":"Reduces the given values with the given accumulator. acc_fun is a function that receives no arguments and returns the actual accumulator. The acc_fun function is invoked per window whenever a new window starts. If a trigger is emitted and it is configured to reset the accumulator, the acc_fun function will be invoked once again. Reducing will accumulate data until a trigger is emitted or until a window completes. When that happens, the returned accumulator will be the new state of the stage and all functions after reduce will be invoked. Examples iex&gt; flow = Flow . from_enumerable ( [ &quot;the quick brown fox&quot; ] ) |&gt; Flow . flat_map ( fn word -&gt; ...&gt; String . graphemes ( word ) ...&gt; end ) iex&gt; flow = flow |&gt; Flow . partition |&gt; Flow . reduce ( fn -&gt; %{ } end , fn grapheme , map -&gt; ...&gt; Map . update ( map , grapheme , 1 , &amp; &amp;1 + 1 ) ...&gt; end ) iex&gt; Enum . sort ( flow ) [ { &quot; &quot; , 3 } , { &quot;b&quot; , 1 } , { &quot;c&quot; , 1 } , { &quot;e&quot; , 1 } , { &quot;f&quot; , 1 } , { &quot;h&quot; , 1 } , { &quot;i&quot; , 1 } , { &quot;k&quot; , 1 } , { &quot;n&quot; , 1 } , { &quot;o&quot; , 2 } , { &quot;q&quot; , 1 } , { &quot;r&quot; , 1 } , { &quot;t&quot; , 1 } , { &quot;u&quot; , 1 } , { &quot;w&quot; , 1 } , { &quot;x&quot; , 1 } ]","ref":"Flow.html#reduce/3","title":"Flow.reduce/3","type":"function"},{"doc":"Applies the given function rejecting each input in parallel. Examples iex&gt; flow = [ 1 , 2 , 3 ] |&gt; Flow . from_enumerable ( ) |&gt; Flow . reject ( &amp; ( rem ( &amp;1 , 2 ) == 0 ) ) iex&gt; Enum . sort ( flow ) # Call sort as we have no order guarantee [ 1 , 3 ]","ref":"Flow.html#reject/2","title":"Flow.reject/2","type":"function"},{"doc":"Runs a given flow. This runs the given flow as a stream for its side-effects. No items are sent from the flow to the current process. Examples iex&gt; parent = self ( ) iex&gt; [ 1 , 2 , 3 ] |&gt; Flow . from_enumerable ( ) |&gt; Flow . map ( &amp; send ( parent , &amp;1 ) ) |&gt; Flow . run ( ) :ok iex&gt; receive do ...&gt; 1 -&gt; :ok ...&gt; end :ok","ref":"Flow.html#run/1","title":"Flow.run/1","type":"function"},{"doc":"Shuffles the data in the given flow (or flows) into a new series of stages with the given window and options. Similar to partition/2 , this function creates a new series of stages to process the data. However, while partition/2 routes the data using the partition dispatcher, shuffle/2 uses GenStage.DemandDispatcher . Examples Flow . shuffle ( flow1 , window : Flow.Window . global ) Flow . shuffle ( [ flow1 , flow2 ] , stages : 4 ) Options :window - a Flow.Window struct which controls how the reducing function behaves, see Flow.Window for more information. :stages - the number of partitions (reducer stages) :shutdown - the shutdown time for this stage when the flow is shut down. The same as the :shutdown value in a Supervisor, defaults to 5000 milliseconds.","ref":"Flow.html#shuffle/2","title":"Flow.shuffle/2","type":"function"},{"doc":"Starts and runs the flow as a separate process. See into_stages/3 in case you want the flow to work as a producer for another series of stages. Options :name - the name of the flow :demand - configures the demand on the flow producers to :forward or :accumulate . The default is :forward . See GenStage.demand/2 for more information. :subscribe_timeout - timeout for the subscription between stages when setting up the flow. Defaults to 5_000 milliseconds. The flow exits with reason :normal only if all consumers exit with reason :normal . Otherwise exits with reason :shutdown .","ref":"Flow.html#start_link/2","title":"Flow.start_link/2","type":"function"},{"doc":"Takes n events according to the sort function. This function allows developers to calculate the top n entries (or the bottom n entries) by performing most of the work concurrently. First n events are taken from every partition and then those n events from every partition are merged into a single partition. The final result is a flow with a single partition that will emit a list with the top n events. The sorting is given by the sort_fun . take_sort/3 is built on top of departition/5 , which means it will also take and sort entries across windows. A set of options may also be given to customize the :window , :min_demand and :max_demand of when departitioning. Examples As an example, imagine you are processing a list of URLs and you want the list of the most accessed URLs. iex&gt; urls = ~w(www.foo.com www.bar.com www.foo.com www.foo.com www.baz.com) iex&gt; flow = urls |&gt; Flow . from_enumerable ( ) |&gt; Flow . partition ( ) iex&gt; flow = flow |&gt; Flow . reduce ( fn -&gt; %{ } end , fn url , map -&gt; ...&gt; Map . update ( map , url , 1 , &amp; &amp;1 + 1 ) ...&gt; end ) iex&gt; flow = flow |&gt; Flow . take_sort ( 1 , fn { _url_a , count_a } , { _url_b , count_b } -&gt; ...&gt; count_b &lt;= count_a ...&gt; end ) iex&gt; Enum . to_list ( flow ) [ [ { &quot;www.foo.com&quot; , 3 } ] ]","ref":"Flow.html#take_sort/4","title":"Flow.take_sort/4","type":"function"},{"doc":"Passes a flow through a list of producer_consumers child specifications and subscriptions that will be started alongside the flow. producers_consumers is a list of tuples where the first element is the child specification and the second is a list of subscription options. The child specification is the one defined in the Supervisor module. The producers_consumers will only be started when the flow starts. If the flow terminates, the producer consumers will also be terminated. The :id field of the child specification will be randomized. The :restart option is set to :temporary but it behaves as :transient . If a producer terminates, its exit reason will propagate through the flow. The exit is considered abnormal unless the reason is :normal , :shutdown or {:shutdown, _} . All other child specification fields are kept unchanged. For options and termination behaviour, see through_stages/3 . Examples spec = { MyConsumerProducer , arg } subscription_opts = [ ] specs = [ { spec , subscription_opts } ] Flow . through_specs ( some_flow , specs )","ref":"Flow.html#through_specs/3","title":"Flow.through_specs/3","type":"function"},{"doc":"Passes a flow through a list of already running stages as producer_consumers . producers_consumers are already running stages that have type :producer_consumer . Each element represents the consumer or a tuple with the consumer and the subscription options as defined in GenStage.sync_subscribe/2 . If instead you want the producer consumers to be started alongside the flow, see through_specs/3 instead. You are required to pass an existing flow and it returns a new flow that you can continue processing. Options These options configure the stages after the producer consumers: :window - a window to run the next stages in, see Flow.Window :stages - the number of stages :buffer_keep - how the buffer should behave, see GenStage.init/1 :buffer_size - how many events to buffer, see GenStage.init/1 :shutdown - the shutdown time for this stage when the flow is shut down. The same as the :shutdown value in a Supervisor, defaults to 5000 milliseconds. All remaining options are sent during subscription, allowing developers to customize :min_demand , :max_demand and others. Examples stages = [ { pid1 , min_demand : 10 } , pid2 , SomeProducerConsumer ] Flow . from_enumerable ( [ 1 , 2 , 3 ] ) |&gt; Flow . through_stages ( stages ) |&gt; Flow . start_link ( ) Termination Flow subscribes to stages using cancel: :transient . This means stages can signal the flow that it has emitted all events by terminating with reason :normal , :shutdown or {:shutdown, _} . If you are implementing your own producer consumer and you are subscribing to a flow that is finite, you need to take this into account in your producer consumer implementation: You need implement GenStage.handle_subscribe/4 and store whenever the stage gets a new producer You need implement GenStage.handle_cancel/3 and decrease whenever the stage loses a producer Once all producers are cancelled, you need to call GenStage.async_info(self(), :terminate) to send a message to yourself, allowing you to terminate after all events have been consumed: def handle_info ( :terminate , state ) do { :stop , :shutdown , state } end Given the complexity in guaranteeing termination, we recommend developers to use through_stages/3 and through_specs/3 only when subscribing to unbounded (infinite) flows. If the exit reason is none of the above, it will cause the next stages to terminate immediately, eventually causing the whole flow to terminate.","ref":"Flow.html#through_stages/3","title":"Flow.through_stages/3","type":"function"},{"doc":"Only emit unique events. Calling this function is equivalent to: Flow . uniq_by ( flow , &amp; &amp;1 ) See uniq_by/2 for more information.","ref":"Flow.html#uniq/1","title":"Flow.uniq/1","type":"function"},{"doc":"Only emit events that are unique according to the by function. In order to verify if an item is unique or not, uniq_by/2 must store the value computed by by/1 into a set. This means that, when working with unbounded data, it is recommended to wrap uniq_by/2 in a window otherwise the data set will grow forever, eventually using all memory available. Also keep in mind that uniq_by/2 is applied per partition. Therefore, if the data is not uniquely divided per partition, it won't be able to calculate the unique items properly. Examples To get started, let's create a flow that emits only the first odd and even number for a range: iex&gt; flow = Flow . from_enumerable ( 1 .. 100 ) iex&gt; flow = Flow . partition ( flow , stages : 1 ) iex&gt; flow |&gt; Flow . uniq_by ( &amp; rem ( &amp;1 , 2 ) ) |&gt; Enum . sort ( ) [ 1 , 2 ] Since we have used only one stage when partitioning, we correctly calculate [1, 2] for the given partition. Let's see what happens when we increase the number of stages in the partition: iex&gt; flow = Flow . from_enumerable ( 1 .. 100 ) iex&gt; flow = Flow . partition ( flow , stages : 4 ) iex&gt; flow |&gt; Flow . uniq_by ( &amp; rem ( &amp;1 , 2 ) ) |&gt; Enum . sort ( ) [ 1 , 2 , 3 , 4 , 10 , 16 , 23 , 39 ] Now we got 8 numbers, one odd and one even per partition . If we want to compute the unique items per partition, we must properly hash the events into two distinct partitions, one for odd numbers and another for even numbers: iex&gt; flow = Flow . from_enumerable ( 1 .. 100 ) iex&gt; flow = Flow . partition ( flow , stages : 2 , hash : fn event -&gt; { event , rem ( event , 2 ) } end ) iex&gt; flow |&gt; Flow . uniq_by ( &amp; rem ( &amp;1 , 2 ) ) |&gt; Enum . sort ( ) [ 1 , 2 ]","ref":"Flow.html#uniq_by/2","title":"Flow.uniq_by/2","type":"function"},{"doc":"Joins two flows with the given window. It is similar to bounded_join/7 with the addition a window can be given. The window function applies to elements of both left and right side in isolation (and not the joined value). A trigger will cause the join state to be cleared. Examples As an example, let's expand the example given in bounded_join/7 and apply a window to it. The example in bounded_join/7 returned 3 results but in this example, because we will split the posts and comments in two different windows, we will get only two results as the later comment for post_id=1 won't have a matching comment for its window: iex&gt; posts = [ %{ id : 1 , title : &quot;hello&quot; , timestamp : 0 } , %{ id : 2 , title : &quot;world&quot; , timestamp : 1000 } ] iex&gt; comments = [ { 1 , &quot;excellent&quot; , 0 } , { 1 , &quot;outstanding&quot; , 1000 } , ...&gt; { 2 , &quot;great follow up&quot; , 1000 } , { 3 , &quot;unknown&quot; , 1000 } ] iex&gt; window = Flow.Window . fixed ( 1 , :second , fn ...&gt; { _ , _ , timestamp } -&gt; timestamp ...&gt; %{ timestamp : timestamp } -&gt; timestamp ...&gt; end ) iex&gt; flow = Flow . window_join ( :inner , ...&gt; Flow . from_enumerable ( posts ) , ...&gt; Flow . from_enumerable ( comments ) , ...&gt; window , ...&gt; &amp; &amp;1 . id , # left key ...&gt; &amp; elem ( &amp;1 , 0 ) , # right key ...&gt; fn post , { _post_id , comment , _ts } -&gt; Map . put ( post , :comment , comment ) end , ...&gt; stages : 1 , max_demand : 1 ) iex&gt; Enum . sort ( flow ) [ %{ id : 1 , title : &quot;hello&quot; , comment : &quot;excellent&quot; , timestamp : 0 } , %{ id : 2 , title : &quot;world&quot; , comment : &quot;great follow up&quot; , timestamp : 1000 } ]","ref":"Flow.html#window_join/8","title":"Flow.window_join/8","type":"function"},{"doc":"","ref":"Flow.html#t:join/0","title":"Flow.join/0","type":"type"},{"doc":"","ref":"Flow.html#t:t/0","title":"Flow.t/0","type":"type"},{"doc":"Splits a flow into windows that are materialized at certain triggers. Windows allow developers to split data so we can understand incoming data as time progresses. Once a window is created, we can specify triggers that allow us to customize when the data accumulated on every window is materialized. Windows must be created by calling one of the window type functions. The supported window types are as follows: Global windows - that's the default window which means all data belongs to one single window. In other words, the data is not split in any way. The window finishes when all producers notify there is no more data Fixed windows - splits incoming events into periodic, non- overlapping windows based on event times. In other words, a given event belongs to a single window. If data arrives late, a configured lateness can be specified. Periodic windows - splits incoming events into periodic, non- overlapping windows based on processing times. Similar to fixed windows, a given event belongs to a single window. Count windows - splits incoming events based on a count. Similar to fixed windows, a given event belongs to a single window. Other common window types can be expressed with Flow functions: Session windows - splits incoming events into unique windows which is grouped until there is a configured gap between event times. Sessions are useful for data that is irregularly distributed with respect to time. We discuss all types and include examples below. In the first section, &quot;Global windows&quot;, we build the basic intuition about windows and triggers as well as discuss the distinction between &quot;Event time and processing time&quot;. Then we explore &quot;Fixed windows&quot; and the concept of lateness before moving on to other window types. Global windows By default, all events belong to the global window. The global window is automatically attached to a partition if no window is specified. The flow below: Flow . from_stages ( [ some_producer ] ) |&gt; Flow . partition ( ) |&gt; Flow . reduce ( fn -&gt; 0 end , &amp; &amp;1 + 2 ) is equivalent to: Flow . from_stages ( [ some_producer ] ) |&gt; Flow . partition ( window : Flow.Window . global ( ) ) |&gt; Flow . reduce ( fn -&gt; 0 end , &amp; &amp;1 + 2 ) Even though the global window does not split the data in any way, it already provides conveniences for working with both bounded (finite) and unbounded (infinite) via triggers. For example, the flow below uses a global window with a count-based trigger to emit the values being summed as we sum them: iex&gt; window = Flow.Window . global ( ) |&gt; Flow.Window . trigger_every ( 10 ) iex&gt; flow = Flow . from_enumerable ( 1 .. 100 ) |&gt; Flow . partition ( window : window , stages : 1 ) iex&gt; flow |&gt; Flow . reduce ( fn -&gt; 0 end , &amp; ( &amp;1 + &amp;2 ) ) |&gt; Flow . emit ( :state ) |&gt; Enum . to_list ( ) [ 55 , 210 , 465 , 820 , 1275 , 1830 , 2485 , 3240 , 4095 , 5050 , 5050 ] Let's explore the types of triggers available next. Triggers Triggers allow us to check point the data processed so far. There are different triggers we can use: Event count triggers - compute state operations every X events Processing time triggers - compute state operations every X time units for every stage Punctuation - hand-written triggers based on the data Flow supports the triggers above via the trigger_every/2 , trigger_periodically/3 and trigger/3 respectively. Once a trigger is emitted, the Flow.reduce/3 step halts and invokes the Flow.on_trigger/2 callback, allowing you to emit events and change the reducer accumulator. Event time and processing time Before we move to other window types, it is important to discuss the distinction between event time and processing time. In particular, triggers created with the trigger_periodically/3 function are intrinsically inaccurate and therefore should not be used to split the data. For example, if you are measuring the frequency that events arrive, using the event time will always yield the same result, while processing time will be vulnerable to fluctuations if, for instance, an external factor causes events to processed slower or faster than usual. Furthermore, periodic triggers are established per partition and are message-based, which means partitions will emit the triggers at different times and possibly with delays based on the partition message queue size. However, it is exactly this lack of precision which makes them efficient for checkpointing data. Flow provides other window types, such as fixed windows, exactly to address the issues with processing time. Such windows use the event time which is based on the data itself. When working with event time, we can assign the data into proper windows even when late or out of order. Such windows can be used to gather time-based insight from the data (for example, the most popular hashtags in the last 10 minutes) as well as for checkpointing data. Fixed windows (event time) Fixed windows group the data based on the event times. Regardless if the data is bounded or not, fixed windows give us time-based insight about the data. Fixed windows are created via the fixed/3 function which specified the duration of the window and a function that retrieves the event time from each event: Flow.Window . fixed ( 1 , :hour , fn { word , timestamp } -&gt; timestamp end ) Let's see an example that will use the window above to count the frequency of words based on windows that are 1 hour long. The timestamps used by Flow are integers in milliseconds. For now, we will also set the concurrency down 1 and max demand down to 5 as it is simpler to reason about the results: iex&gt; data = [ { &quot;elixir&quot; , 0 } , { &quot;elixir&quot; , 1_000 } , { &quot;erlang&quot; , 60_000 } , ...&gt; { &quot;concurrency&quot; , 3_200_000 } , { &quot;elixir&quot; , 4_000_000 } , ...&gt; { &quot;erlang&quot; , 5_000_000 } , { &quot;erlang&quot; , 6_000_000 } ] iex&gt; window = Flow.Window . fixed ( 1 , :hour , fn { _word , timestamp } -&gt; timestamp end ) iex&gt; flow = Flow . from_enumerable ( data , max_demand : 5 , stages : 1 ) iex&gt; flow = Flow . partition ( flow , window : window , stages : 1 ) iex&gt; flow = Flow . reduce ( flow , fn -&gt; %{ } end , fn { word , _ } , acc -&gt; ...&gt; Map . update ( acc , word , 1 , &amp; &amp;1 + 1 ) ...&gt; end ) iex&gt; flow |&gt; Flow . emit ( :state ) |&gt; Enum . to_list [ %{ &quot;elixir&quot; =&gt; 2 , &quot;erlang&quot; =&gt; 1 , &quot;concurrency&quot; =&gt; 1 } , %{ &quot;elixir&quot; =&gt; 1 , &quot;erlang&quot; =&gt; 2 } ] Since the data has been broken in two windows, the first four events belong to the same window while the last 3 belongs to the second one. Notice that Flow.reduce/3 is executed per window and that each event belongs to a single window exclusively. Similar to global windows, fixed windows can also have triggers, allowing us to checkpoint the data as the computation happens. Data ordering, watermarks and lateness When working with event time, Flow assumes by default that events are time ordered. This means that, when we move from one window to another, like when we received the entry {&quot;elixir&quot;, 4_000_000} in the example above, we assume the previous window has been completed. Let's change the events above to be out of order and move the first event to the end of the dataset and see what happens: iex&gt; data = [ { &quot;elixir&quot; , 1_000 } , { &quot;erlang&quot; , 60_000 } , ...&gt; { &quot;concurrency&quot; , 3_200_000 } , { &quot;elixir&quot; , 4_000_000 } , ...&gt; { &quot;erlang&quot; , 5_000_000 } , { &quot;erlang&quot; , 6_000_000 } , { &quot;elixir&quot; , 0 } ] iex&gt; window = Flow.Window . fixed ( 1 , :hour , fn { _word , timestamp } -&gt; timestamp end ) iex&gt; flow = Flow . from_enumerable ( data ) |&gt; Flow . partition ( window : window , stages : 1 , max_demand : 5 ) iex&gt; flow = Flow . reduce ( flow , fn -&gt; %{ } end , fn { word , _ } , acc -&gt; ...&gt; Map . update ( acc , word , 1 , &amp; &amp;1 + 1 ) ...&gt; end ) iex&gt; flow |&gt; Flow . emit ( :state ) |&gt; Enum . to_list [ %{ &quot;elixir&quot; =&gt; 1 , &quot;erlang&quot; =&gt; 1 , &quot;concurrency&quot; =&gt; 1 } , %{ &quot;elixir&quot; =&gt; 1 , &quot;erlang&quot; =&gt; 2 } ] Notice that now the first map did not count the &quot;elixir&quot; word twice. Since the event arrived late, it was marked as lost. However, in many flows we actually expect data to arrive late or out of order, especially when talking about concurrent data processing. Luckily, event time windows include the concept of lateness, which is a processing time base period we would wait to receive late events. Let's change the example above once more but now change the window to also call allowed_lateness/3 : iex&gt; data = [ { &quot;elixir&quot; , 1_000 } , { &quot;erlang&quot; , 60_000 } , ...&gt; { &quot;concurrency&quot; , 3_200_000 } , { &quot;elixir&quot; , 4_000_000 } , ...&gt; { &quot;erlang&quot; , 5_000_000 } , { &quot;erlang&quot; , 6_000_000 } , { &quot;elixir&quot; , 0 } ] iex&gt; window = Flow.Window . fixed ( 1 , :hour , fn { _word , timestamp } -&gt; timestamp end ) iex&gt; window = Flow.Window . allowed_lateness ( window , 5 , :minute ) iex&gt; flow = Flow . from_enumerable ( data ) |&gt; Flow . partition ( window : window , stages : 1 , max_demand : 5 ) iex&gt; flow = Flow . reduce ( flow , fn -&gt; %{ } end , fn { word , _ } , acc -&gt; ...&gt; Map . update ( acc , word , 1 , &amp; &amp;1 + 1 ) ...&gt; end ) iex&gt; flow |&gt; Flow . emit ( :state ) |&gt; Enum . to_list [ %{ &quot;concurrency&quot; =&gt; 1 , &quot;elixir&quot; =&gt; 1 , &quot;erlang&quot; =&gt; 1 } , %{ &quot;concurrency&quot; =&gt; 1 , &quot;elixir&quot; =&gt; 2 , &quot;erlang&quot; =&gt; 1 } , %{ &quot;elixir&quot; =&gt; 1 , &quot;erlang&quot; =&gt; 2 } ] Now that we allow late events, we can see the first window emitted twice. Instead of the window being marked as done when 1 hour passes, we say it emits a watermark trigger . The window will be effectively done only after the allowed lateness period. If desired, we can use Flow.on_trigger/2 to get more information about each particular window and its trigger. Replace the last line above by the following: flow |&gt; Flow . on_trigger ( fn state , _index , trigger -&gt; { [ { state , trigger } ] , state } end ) |&gt; Enum . to_list ( ) The trigger parameter will include the type of window, the current window and what caused the window to be emitted ( :watermark or :done ). Note that all stages must receive an event that is outside of a specific window before that window is considered complete. In other words if there are multiple stages in the partition preceding a reduce operation that has a window, the reduce step won't release a window until it has seen an event that is outside of that window from all processes that it receives data from. This could have an effect on how long events are delayed in the reduce step. Periodic windows (processing time) Periodic windows are similar to fixed windows except triggers are emitted based on processing time instead of event time. Remember that relying on periodic windows or triggers is intrinsically inaccurate and should not be used to split the data, only as a checkpointing device. Periodic windows are also similar to global windows that use trigger_periodically/2 to emit events periodically. The difference is that periodic windows emit a window in a given interval while a trigger emits a trigger. This behaviour may affect functions such as Flow.departition/4 , which calls the merge callback per trigger but the done callback per window. Unless you are relying on functions such as Flow.departition/4 , there is no distinction between periodic windows and global windows with periodic triggers. Count windows (event count) Count windows are simpler versions of fixed windows where windows are split apart by event count. Since it is not timed-based, it does not provide the concept of lateness. iex&gt; window = Flow.Window . count ( 10 ) iex&gt; flow = Flow . from_enumerable ( 1 .. 100 ) |&gt; Flow . partition ( window : window , stages : 1 ) iex&gt; flow |&gt; Flow . reduce ( fn -&gt; 0 end , &amp; ( &amp;1 + &amp;2 ) ) |&gt; Flow . emit ( :state ) |&gt; Enum . to_list ( ) [ 55 , 155 , 255 , 355 , 455 , 555 , 655 , 755 , 855 , 955 , 0 ] Count windows are also similar to global windows that use trigger_every/2 to emit events per count. The difference is that count windows emit a window per event count while a trigger belongs to a window. This behaviour may affect functions such as Flow.departition/4 , which calls the merge callback per trigger but the done callback per window. Unless you are relying on functions such as Flow.departition/4 , there is no distinction between count windows and global windows with count triggers. Session windows (gap between events) Session windows allow events to accumulate until a configured time gap between events occurs. This allows for grouping events that occurred close to each other, while allowing the length of the window to vary. Flow does not provide a dedicated Session window type, but it can be constructed using emit_and_reduce/3 and on_trigger/2 . iex&gt; data = [ ...&gt; { &quot;elixir&quot; , 2_000_000 } , ...&gt; { &quot;erlang&quot; , 3_100_000 } , ...&gt; { &quot;elixir&quot; , 3_200_000 } , ...&gt; { &quot;erlang&quot; , 4_000_000 } , ...&gt; { &quot;elixir&quot; , 4_100_000 } , ...&gt; { &quot;erlang&quot; , 4_150_000 } ...&gt; ] iex&gt; max_gap_between_events = 1_000_000 iex&gt; flow = Flow . from_enumerable ( data ) |&gt; Flow . partition ( key : fn { k , _ } -&gt; k end , stages : 1 ) iex&gt; flow = ...&gt; Flow . emit_and_reduce ( flow , fn -&gt; %{ } end , fn { word , time } , acc -&gt; ...&gt; { count , previous_time } = Map . get ( acc , word , { 1 , time } ) ...&gt; ...&gt; if time - previous_time &gt; max_gap_between_events do ...&gt; { [ { word , { count , previous_time } } ] , Map . put ( acc , word , { 1 , time } ) } ...&gt; else ...&gt; { [ ] , Map . update ( acc , word , { 1 , time } , fn { count , _ } -&gt; { count + 1 , time } end ) } ...&gt; end ...&gt; end ) iex&gt; flow = Flow . on_trigger ( flow , fn acc -&gt; { Enum . to_list ( acc ) , :unused } end ) iex&gt; Enum . to_list ( flow ) [ { &quot;elixir&quot; , { 1 , 2000000 } } , { &quot;elixir&quot; , { 2 , 4100000 } } , { &quot;erlang&quot; , { 3 , 4150000 } } ]","ref":"Flow.Window.html","title":"Flow.Window","type":"module"},{"doc":"Sets a duration, in processing time, of how long we will wait for late events for a given window. If allowed lateness is configured, once the window is finished, it won't trigger a :done event but instead emit a :watermark . The window will be done only when the allowed lateness time expires, effectively emitting the :done trigger. count is a positive number. The unit may be a time unit ( :millisecond , :second , :minute , or :hour ).","ref":"Flow.Window.html#allowed_lateness/3","title":"Flow.Window.allowed_lateness/3","type":"function"},{"doc":"Returns a count-based window of every count elements. count must be a positive integer. Count window triggers have the shape of {:count, window, trigger_name} , where window is an incrementing integer identifying the window. See the section on &quot;Count windows&quot; in the module documentation for examples.","ref":"Flow.Window.html#count/1","title":"Flow.Window.count/1","type":"function"},{"doc":"Returns a fixed window of duration count unit where the event time is calculated by the given function by . count is a positive integer and unit is one of :millisecond , :second , :minute , or :hour . Fixed window triggers have the shape of {:fixed, window, trigger_name} , where window is an integer that represents the beginning timestamp for the current window. If allowed_lateness/3 is used with fixed windows, the window will first emit a {:fixed, window, :watermark} trigger when the window terminates and emit {:fixed, window, :done} only after the allowed_lateness/3 duration has passed. See the section on &quot;Fixed windows&quot; in the module documentation for examples.","ref":"Flow.Window.html#fixed/3","title":"Flow.Window.fixed/3","type":"function"},{"doc":"Returns a global window. Global window triggers have the shape of {:global, :global, trigger_name} . See the section on &quot;Global windows&quot; in the module documentation for examples.","ref":"Flow.Window.html#global/0","title":"Flow.Window.global/0","type":"function"},{"doc":"Returns a period-based window of every count unit . count is a positive integer and unit is one of :millisecond , :second , :minute , or :hour . Remember periodic triggers are established per partition and are message-based, which means partitions will emit the triggers at different times and possibly with delays based on the partition message queue size. Periodic window triggers have the shape of {:periodic, window, trigger_name} , where window is an incrementing integer identifying the window. See the section on &quot;Periodic windows&quot; in the module documentation for examples.","ref":"Flow.Window.html#periodic/2","title":"Flow.Window.periodic/2","type":"function"},{"doc":"Calculates when to emit a trigger. Triggers are calculated per window and are used to temporarily halt the window accumulation, typically done with Flow.reduce/3 , allowing the next operations to execute before accumulation is resumed. This function expects the trigger accumulator function, which will be invoked at the beginning of every window, and a trigger function that receives the current batch of events and its own accumulator. The trigger function must return one of the three values: {:cont, acc} - the reduce operation should continue as usual. acc is the trigger state. {:cont, events, acc} - the reduce operation should continue, but only with the events you want to emit as part of the next state. acc is the trigger state. {:trigger, name, pre, pos, acc} - where name is the trigger name , pre are the events to be consumed before the trigger, pos controls events to be processed after the trigger with the acc as the new trigger accumulator. We recommend looking at the implementation of trigger_every/2 as an example of a custom trigger.","ref":"Flow.Window.html#trigger/3","title":"Flow.Window.trigger/3","type":"function"},{"doc":"A trigger emitted every count elements in a window. The trigger will be named {:every, count} . Examples Below is an example that checkpoints the sum from 1 to 100, emitting a trigger with the state every 10 items. The extra 5050 value at the end is the trigger emitted because processing is done. iex&gt; window = Flow.Window . global ( ) |&gt; Flow.Window . trigger_every ( 10 ) iex&gt; flow = Flow . from_enumerable ( 1 .. 100 ) |&gt; Flow . partition ( window : window , stages : 1 ) iex&gt; flow |&gt; Flow . reduce ( fn -&gt; 0 end , &amp; ( &amp;1 + &amp;2 ) ) |&gt; Flow . emit ( :state ) |&gt; Enum . to_list ( ) [ 55 , 210 , 465 , 820 , 1275 , 1830 , 2485 , 3240 , 4095 , 5050 , 5050 ]","ref":"Flow.Window.html#trigger_every/2","title":"Flow.Window.trigger_every/2","type":"function"},{"doc":"Emits a trigger periodically every count unit . Such trigger will apply to every window that has changed since the last periodic trigger. count is a positive integer and unit is one of :millisecond , :second , :minute , or :hour . Remember periodic triggers are established per partition and are message-based, which means partitions will emit the triggers at different times and possibly with delays based on the partition message queue size. The trigger will be named {:periodically, count, unit} . Message-based triggers (timers) It is also possible to dispatch a trigger by sending a message to self() with the format of {:trigger, name} . This is useful for custom triggers and timers. One example is to send the message when building the accumulator for Flow.reduce/3 . Similar to periodic triggers, message-based triggers will also be invoked to all windows that have changed since the last trigger.","ref":"Flow.Window.html#trigger_periodically/3","title":"Flow.Window.trigger_periodically/3","type":"function"},{"doc":"A function that returns the event time to window by. It must return an integer representing the time in milliseconds. Flow does not care if the integer is using the UNIX epoch, Gregorian epoch or any other as long as it is consistent.","ref":"Flow.Window.html#t:by/0","title":"Flow.Window.by/0","type":"type"},{"doc":"The window identifier. It is :global for :global windows or an integer for fixed windows.","ref":"Flow.Window.html#t:id/0","title":"Flow.Window.id/0","type":"type"},{"doc":"","ref":"Flow.Window.html#t:t/0","title":"Flow.Window.t/0","type":"type"},{"doc":"The supported time units for fixed and periodic windows.","ref":"Flow.Window.html#t:time_unit/0","title":"Flow.Window.time_unit/0","type":"type"},{"doc":"The name of the trigger.","ref":"Flow.Window.html#t:trigger/0","title":"Flow.Window.trigger/0","type":"type"},{"doc":"The supported window types.","ref":"Flow.Window.html#t:type/0","title":"Flow.Window.type/0","type":"type"}]